"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
exports.Model = void 0;
var qs = require("qs");
var methods_1 = require("./methods");
var model_1 = require("./model");
exports.Model = model_1.default;
var request_1 = require("./request");
var utils_1 = require("./utils");
var INSTANCES = {};
function resolve(key) {
    if (!INSTANCES[key]) {
        INSTANCES[key] = create();
    }
    return INSTANCES[key];
}
function create(options) {
    var client = function client(path) {
        var _a;
        return _a = (function (_super) {
                __extends(AnonymousModel, _super);
                function AnonymousModel() {
                    return _super !== null && _super.apply(this, arguments) || this;
                }
                return AnonymousModel;
            }(model_1.default)),
            _a.path = path,
            _a.client = client,
            _a;
    };
    client._options = options || {};
    client._count = 0;
    client._tasks = [];
    client.create = create;
    client.resolve = resolve;
    client.setOptions = function (opts) {
        client._options = Object.assign({}, client._options, opts);
    };
    client.getFormDataClass = function getFormDataClass() {
        var FormData = client._options.FormData;
        if (!FormData && typeof window === 'object' && window.FormData) {
            FormData = window.FormData;
        }
        else if (!FormData && typeof global === 'object' && global.FormData) {
            FormData = global.FormData;
        }
        return FormData;
    };
    client.createBody = function (body) {
        var FormData = client.getFormDataClass();
        if (!body || typeof body !== 'object' || !FormData || utils_1.isUint8Array(body) || body instanceof FormData)
            return body;
        var form = null;
        function addField(path, value) {
            if (value === undefined)
                return;
            if (!utils_1.isReadableStream(value) && !utils_1.isFile(value) && !utils_1.isUint8Array(value)) {
                if (typeof value === 'boolean' || typeof value === 'number' || value === null || value instanceof Date) {
                    value = String(value);
                }
                else if (typeof value === 'object') {
                    for (var key in value) {
                        addField(path + "[" + key + "]", value[key]);
                    }
                    return;
                }
            }
            form.append(path, value);
        }
        for (var key in body) {
            var value = body[key];
            if (utils_1.isReadableStream(value) || utils_1.isFile(value) || utils_1.isUint8Array(value)) {
                form = new FormData();
                Object.keys(body).forEach(function (path) { return addField(path, body[path]); });
                break;
            }
        }
        return form || body;
    };
    client.request = function request(path, init, query, reducer) {
        init = Object.assign({}, init);
        var queryParams = Object.assign({}, init.query);
        delete init.query;
        var queryString = qs.stringify(queryParams, client._options.qsOptions);
        if (queryString) {
            if (path.indexOf('?') < 0) {
                queryString = "?" + queryString;
            }
            else if (path[path.length - 1] !== '&' && path[path.length - 1] !== '?') {
                queryString = "&" + queryString;
            }
            path += queryString;
        }
        var apiRoot = client._options.apiRoot;
        if (apiRoot) {
            if (apiRoot[apiRoot.length - 1] === '/' && path[0] === '/') {
                path = path.substring(1);
            }
            else if (apiRoot[apiRoot.length - 1] !== '/' && path[0] !== '/') {
                path = "/" + path;
            }
            path = apiRoot + path;
        }
        var fetch = client._options.fetch;
        if (!fetch) {
            if (typeof window !== 'undefined') {
                fetch = window.fetch.bind(window);
            }
            else if (typeof global !== 'undefined') {
                fetch = global.fetch;
            }
        }
        var req = new request_1.default(client, fetch, path, init, query, reducer);
        client._count += 1;
        client._tasks.push(req);
        client._updateProgress();
        return req;
    };
    client._updateProgress = function () {
        if (client._updateProgressTimer || !client._options.onProgress)
            return;
        client._updateProgressTimer = setTimeout(function () {
            client._updateProgressTimer = 0;
            if (!client._options.onProgress)
                return;
            var progress = 1;
            var total = client._tasks.length * 3;
            if (total) {
                var steps_1 = 0;
                client._tasks.forEach(function (r) {
                    steps_1 += r._steps;
                });
                progress = steps_1 / total;
            }
            client._options.onProgress(progress);
        }, 5);
    };
    methods_1.default.forEach(function (method) {
        client[method] = function (path, init) {
            init = init || {};
            init.method = method.toUpperCase();
            return client.request(path, init);
        };
    });
    return client;
}
exports.default = resolve('default');
